package tinyvm

import java.time.Instant
import kotlin.test.Test
import kotlin.test.assertEquals

internal class RepositoryTest {
    private val repository = Repository()
    private val commits =
        listOf(
            Commit(
                tree = Tree(mapOf("test1.txt" to Blob("Hello World!"))),
                author = Author("Gleb Koval", "gleb@koval.net"),
                message = "Add test1.txt",
                timestamp = Instant.ofEpochSecond(0),
            ),
            Commit(
                tree =
                    Tree(
                        mapOf(
                            "dir2" to
                                Tree(
                                    mapOf(
                                        "test1.txt" to Blob("This is a second file"),
                                    ),
                                ),
                            "dir1" to
                                Tree(
                                    mapOf(
                                        "test1.txt" to Blob("Hello World!"),
                                    ),
                                ),
                        ),
                    ),
                author = Author("Gleb Koval", "gleb@koval.net"),
                message = "Move test1.txt to dir1 and add dir2/test1.txt",
                timestamp = Instant.ofEpochSecond(50),
            ),
            Commit(
                tree =
                    Tree(
                        mapOf(
                            "dir1" to
                                Tree(
                                    mapOf(
                                        "test1.txt" to Blob("Hello World!"),
                                    ),
                                ),
                            "dir2" to
                                Tree(
                                    mapOf(
                                        "test1.txt" to Blob("This is a second file"),
                                    ),
                                ),
                            "README.md" to Blob("# This is a test repo!"),
                        ),
                    ),
                author = Author("Gleb Koval", "gleb@koval.net"),
                message = "Add README.md",
                timestamp = Instant.ofEpochSecond(100),
            ),
        )

    @Test
    fun `can commit`() {
        val committed = repository.commit(commits[0])
        assertEquals(commits[0].hash(), committed.hash())
        assertEquals(1, repository.listCommits().size)
        assertEquals(3, repository.dumpObjects().size)
    }

    @Test
    fun `can deduplicate commit`() {
        repository.commit(commits[0])
        assertEquals(1, repository.listCommits().size)
        val committed = repository.commit(commits[0])
        assertEquals(1, repository.listCommits().size)
        assertEquals(3, repository.dumpObjects().size)
        assertEquals(commits[0].hash(), committed.hash())
    }

    @Test
    fun `can commit twice and deduplicate objects`() {
        repository.commit(commits[0])
        val committed = repository.commit(commits[1])
        assertEquals(2, repository.listCommits().size)
        assertEquals(7, repository.dumpObjects().size)
        assertEquals(commits[1].hash(), committed.hash())
        assertEquals(commits.take(2).map { it.hash() }, repository.listCommits().map { it.hash() })
    }

    @Test
    fun `can commit in-between existing commits`() {
        repository.commit(commits[0])
        repository.commit(commits[2])
        val committed = repository.commit(commits[1])
        assertEquals(3, repository.listCommits().size)
        assertEquals(10, repository.dumpObjects().size)
        assertEquals(commits[1].hash(), committed.hash())
        assertEquals(commits.map { it.hash() }, repository.listCommits().map { it.hash() })
    }

    @Test
    fun `can get commit by hash`() {
        repository.commit(commits[0])
        val committed = repository.commit(commits[1])
        repository.commit(commits[2])
        assertEquals(committed, repository.getCommit(commits[1].hash()))
        assertEquals(null, repository.getCommit("00000000000000000000"))
    }

    @Test
    fun `can find commit by predicate`() {
        repository.commit(commits[0])
        val committed = repository.commit(commits[1])
        repository.commit(commits[2])
        assertEquals(committed, repository.findCommit { it.message.matches("Move.*".toRegex()) })
        assertEquals(null, repository.getCommit("00000000000000000000"))
    }

    @Test
    fun `can create and use branch`() {
        repository.commit(commits[0])
        repository.commit(commits[1])
        repository.commit(commits[2])
        repository.createBranch("test-branch", use = false)
        assertEquals("master", repository.branch)
        repository.createBranch("test-branch-2")
        assertEquals("test-branch-2", repository.branch)
        assertEquals(setOf("master", "test-branch", "test-branch-2"), repository.branches)
        assertEquals(commits.map { it.hash() }, repository.listCommits().map { it.hash() })
        assertEquals(10, repository.dumpObjects().size)
    }

    @Test
    fun `can commit to branch`() {
        repository.commit(commits[0])
        repository.createBranch("test-branch")
        repository.commit(commits[1])
        repository.createBranch("test-branch-2")
        repository.commit(commits[2])

        assertEquals(commits.map { it.hash() }, repository.listCommits().map { it.hash() })
        repository.useBranch("test-branch")
        assertEquals(commits.take(2).map { it.hash() }, repository.listCommits().map { it.hash() })
        repository.useBranch("master")
        assertEquals(commits.take(1).map { it.hash() }, repository.listCommits().map { it.hash() })
    }

    @Test
    fun `can display commit`() {
        assertEquals(
            """
            commit 804c6f0c66da0ea0eab8ac29f23a627e03a962b2
            tree aaebbb258bce30749fc302cbd161b78462252a32
            author Gleb Koval <gleb@koval.net>
            timestamp 0
            
            Add test1.txt
            """.trimIndent(),
            commits[0].toString(),
        )
    }

    @Test
    fun `can display tree`() {
        assertEquals(
            """
            tree 048b8e267dd7a6c5d4849d59f6ff82e6ae948f13
            +dir1/
            | tree aaebbb258bce30749fc302cbd161b78462252a32
            | +test1.txt
            | | blob c57eff55ebc0c54973903af5f72bac72762cf4f4
            | | | Hello World!
            +dir2/
            | tree 0e14cbdba02924cd63a58c2492fc3dd05dc682bd
            | +test1.txt
            | | blob da3fcc31cbc16fcbb7526d68771f77e7e7f02fb1
            | | | This is a second file
            """.trimIndent(),
            commits[1].tree.toString(),
        )
    }
}